//
//  PSMWebSpiderPlugin.m
//  Pandora
//
//  Created by John Pannell on 2/16/06.
//  Copyright 2006 Positive Spin Media. All rights reserved.
//

#import "PSMWebSpiderPlugin.h"

@interface PSMWebSpiderPlugin (Private)
- (NSMutableDictionary *)preferences;
@end

@implementation PSMWebSpiderPlugin

#pragma mark -
#pragma mark Creation/Destruction

+ (void)initialize
{
    // if preferences do not yet exist, I must register some defaults
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    if([[defaults persistentDomainNames] containsObject:[[NSBundle bundleForClass:[self class]] bundleIdentifier]]){
        return; // already exists
    }

    // create dictionary
    NSMutableDictionary *defaultValues = [NSMutableDictionary dictionary];
    
    // set values
    [defaultValues setObject:[NSNumber numberWithBool:NO] forKey:@"stay within domain"];
    [defaultValues setObject:[NSNumber numberWithBool:NO] forKey:@"stay within subdirectory"];
    [defaultValues setObject:[NSNumber numberWithBool:NO] forKey:@"observe link depth limit"];
    [defaultValues setObject:[NSNumber numberWithInt:1] forKey:@"link depth limit"];
    
    // write to disk
    [defaults setPersistentDomain:defaultValues forName:[[NSBundle bundleForClass:[self class]] bundleIdentifier]];
}

- (id)init
{
    if(self = [super initWithWindowNibName:@"BasicSpiderConfig"]){
        _maxFileSize = 0;
        _minFileSize = 0;
        _preferences = [[[NSUserDefaults standardUserDefaults] persistentDomainForName:[[NSBundle bundleForClass:[self class]] bundleIdentifier]] mutableCopy];
    }
    return self;
}

- (void)awakeFromNib
{
    [icon setImage:[[self class] icon]];
}

#pragma mark -
#pragma mark Identification

+ (NSImage *)icon
{
    return [[NSWorkspace sharedWorkspace] iconForFile:[[NSBundle bundleForClass:[self class]] bundlePath]];
}

+ (NSString *)name
{
    return [[[[NSBundle bundleForClass:[self class]] bundlePath] lastPathComponent] stringByDeletingPathExtension];
}

+ (NSString *)description
{
    return @"Finds files by spidering out from a given URL.";
}

+ (NSString *)category
{
    return @"Spiders";
}

+ (NSString *)prompt
{
    return @"Type a URL here";
}

#pragma mark -
#pragma mark Accessors

- (NSMutableDictionary *)preferences
{
    return _preferences;
}

- (void)setMIMETypes:(NSSet *)types
{
    [types retain];
    [_MIMETypes release];
    _MIMETypes = types;
}

- (void)setMinFileSize:(int)kilobytes
{
    _minFileSize = kilobytes;
}

- (void)setMaxFileSize:(int)kilobytes
{
    _maxFileSize = kilobytes;
}

#pragma mark -
#pragma mark Preferences

- (IBAction)recordPreferences:(id)sender
{
    // end editing for link depth field so value is picked up with bindings
    [[linkDepthField cell] endEditing:[[self window] fieldEditor:NO forObject:linkDepthField]];
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    [defaults setPersistentDomain:_preferences forName:[[NSBundle bundleForClass:[self class]] bundleIdentifier]];
    [defaults synchronize];
    [NSApp endSheet:[self window]];
    [[self window] orderOut:self];
    [NSApp stopModal];
    [self release];
}

#pragma mark -
#pragma mark Spider

- (BOOL)contentFromDifferentServers
{
    // return YES if your plugin rewrites returned links to fetch varous contents from different servers, otherwise return NO.  This would only be YES if you were scraping search engine result pages and returning links to source content.  It should definitely be NO for a spider that moves from link to link on the same server.
    return NO;
}

- (int)linkDepth
{
    // you can keep the spider limited to a specific link depth if desired.  Return -1 if there is no restriction.
    if(![[_preferences objectForKey:@"observe link depth limit"] boolValue])
        return -1;
    
    return [[_preferences objectForKey:@"link depth limit"] intValue];
}

- (NSMutableArray *)preprocessURLOrKeyword:(NSString *)string
{
    // identify server
    NSURL *testURL;
    NSMutableString *targetURL = [string mutableCopy];
    
    if([string rangeOfString:@"://"].length == 0){
        // no scheme, add default
        [targetURL insertString:@"http://" atIndex:0];
    }

    testURL = [NSURL URLWithString:targetURL];
    if(testURL == nil){
        NSLog(@"Could not form valid URL with entered information");
    }
    
    _server = [[testURL host] retain];
    _subdirectory = [[NSString stringWithFormat:@"%@%@", _server, [[testURL path] stringByDeletingLastPathComponent]] retain];
            
    NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:targetURL, @"URL", @"parse", @"action", nil];
    NSMutableArray *returnArray = [NSMutableArray arrayWithObject:dict];
    
    return returnArray;
}

- (BOOL)shouldPursueResourceWithURL:(NSString *)url MIMEType:(NSString *)mime andExpectedFileSize:(long long)size withReason:(NSString **)reason
{
    if(![_MIMETypes containsObject:mime]){
        *reason = @"Wrong MIME Type";
        return NO;
    }
    
    if((_minFileSize != 0) && (size < (_minFileSize * 1024))){
        *reason = @"File is too small";
        return NO;
    }
    
    if((_maxFileSize != 0) && (size > (_maxFileSize * 1024))){
        *reason = @"File is too large";
        return NO;
    }
    
    return YES;
}

- (NSDictionary *)postprocessURL:(NSString *)url
{
    // given a URL, you can determine not to use it (return nil), to alter it (return another string)
    // or pass it back unchanged.  The URLs are fed to this function right after they are discovered on
    // a parsed HTML web page.  You must package the returned string into a dictionary... the dictionary
    // currently supports two keys: source and parent - both should be strings representing URLs.  The
    // source key is for the URL that should be fetched, and the "parent" is metadata, so the file knows
    // what web page it came from.  The parent key is optional, and you would only know it in the case where
    // you are scraping a search engine or something.
    if([[_preferences objectForKey:@"stay within domain"] boolValue] && [url rangeOfString:_server].length == 0)
        return nil;
    
    if([[_preferences objectForKey:@"stay within subdirectory"] boolValue] && [url rangeOfString:_subdirectory].length == 0)
        return nil;
    
    return [NSDictionary dictionaryWithObjectsAndKeys:url, @"source", nil];
}

#pragma mark -
#pragma mark Authentication

- (NSArray *)authenticationInfo
{
    // the return array has two objects, an array of keys, and an array of corresponding values.
    // these keys and values are used to fill out a web form and submit it, or to 
    // provide to HTTP authentication.
    
    // For web forms, the keys are the names of the input elements that need to be filled out.
    // note that the keys and values are case sensitive!
    
    // if providing info for HTTP authentication, provide keys "username" and "password", in that order.
    
    // return [NSArray arrayWithObjects:[NSArray arrayWithObjects:@"username", @"password", nil], [NSArray arrayWithObjects:@"myUsername", @"myPassword", nil], nil];
    
    return nil;
}

- (BOOL)parseAuthSuccessPage
{
    return NO;
}

- (NSString *)authSuccessURLFragment
{
    return nil;
}

@end
