//
//  PSMGLImage.m
//  NetScrape
//
//  Created by John Pannell on Sat Jan 01 2005.
//  Copyright (c) 2005 Positive Spin Media. All rights reserved.
//
//  Heavily leveraged from Apple's GLImage example code

/*
    desired behavior: PSMGLImage has a bitmap textured onto a GL quad.  The quad has the same aspect
    ratio as the view, and the bitmap can be scaled to fit in two different ways: proportional
    (entire image just fits) or scale to fit (all of screen is occupied, some of image is cut off).
 
    The major (longer) axis of the view is scaled to a value of 100.0f, the minor axis is less.
 */

#import "PSMGLImage.h"
#import "GLString.h"

@implementation PSMGLImage

-(NSBitmapImageRep*)bmpImage
{
    return bmpRep;
}

-(GLString*)imageInfo
{
    return imageInfo;
}

-(float)zOffset
{
    return zOffset;
}

-(int)txID
{
    return txID;
}

-(id)initWithFrame:(NSRect)frame
{
    self = [super init];
    viewWidth = frame.size.width;
    viewHeight = frame.size.height;
    viewRatio = (float)viewWidth/(float)viewHeight;
    // scale for gl coordinates
    if(viewRatio > 1.0){
        // wider
        glViewWidth = 100.0f;
        glViewHeight = 100.0f / viewRatio;
    } else {
        // taller
        glViewHeight = 100.0f;
        glViewWidth = 100.0f * viewRatio;
    }
    txID = 0;
    myAlpha = 1.0;
    ready = NO;
    return self;
}

-(void)setFrame:(NSRect)frame
{
    // frame is resized from view
    viewWidth = frame.size.width;
    viewHeight = frame.size.height;
    viewRatio = (float)viewWidth/(float)viewHeight;
    // scale for gl coordinates
    if(viewRatio > 1.0){
        // wider
        glViewWidth = 100.0f;
        glViewHeight = 100.0f / viewRatio;
    } else {
        // taller
        glViewHeight = 100.0f;
        glViewWidth = 100.0f * viewRatio;
    }
}

-(void)genQuad
{
    // generate a quad based on full view aspect ratio and image size
    // major axis is 100.0f in glCoordinate
    
    float scaledX, scaledY, scaledZ;
    
    // depth should be 1/2 the length of the dimension perpendicular to rotation
    switch([[NSUserDefaults standardUserDefaults] integerForKey:@"Slideshow secondary transition"]){
        case 0:
        case 1:
            // vertical roll
            scaledZ = -0.5 * glViewHeight;
            break;
            
        case 2:
        case 3:
            // horizontal roll
            scaledZ = -0.5 * glViewWidth;
            break;
            
        default:
            // should never be called
            scaledZ = -0.5 * glViewHeight;
            break;
    }
    // if the transition is a flip, z is zero.
    if([[NSUserDefaults standardUserDefaults] integerForKey:@"Slideshow primary transition"] == 2){
        scaledZ = 0.0;
    }
    // save for view to ask about
    zOffset = scaledZ;
    
    // background quad generation
    scaledX = 0.5 * glViewWidth;
    scaledY = 0.5 * glViewHeight;
    backgroundQuad = malloc(sizeof(Vector3) * 4);
    
    backgroundQuad[0].x = scaledX;
    backgroundQuad[0].y = scaledY;
    backgroundQuad[0].z = scaledZ;
    
    backgroundQuad[1].x = -scaledX;
    backgroundQuad[1].y = scaledY;
    backgroundQuad[1].z = scaledZ;
    
    backgroundQuad[2].x = -scaledX;
    backgroundQuad[2].y = -scaledY;
    backgroundQuad[2].z = scaledZ;
    
    backgroundQuad[3].x = scaledX;
    backgroundQuad[3].y = -scaledY;
    backgroundQuad[3].z = scaledZ;
    
    
    // determine image glcoordinates from image size and glview size
    if([[NSUserDefaults standardUserDefaults] boolForKey:@"Slideshow scale images"]){
        // images are to fill the view
        scaledX = 0.5 * glViewWidth;
        scaledY = 0.5 * glViewHeight;
    } else {
        // do not fill the view
        if(viewRatio > imgRatio){
            // image limited vertically - will not fill sides
            scaledX = 0.5 * glViewWidth * (imgRatio/viewRatio);
            scaledY = 0.5 * glViewHeight;
        } else {
            // image limited horizontally - will not fill top/bottom
            scaledX = 0.5 * glViewWidth;
            scaledY = 0.5 * glViewHeight * (viewRatio/imgRatio);
        }
    }
    
    glQuad = malloc(sizeof(Vector3) * 4);
    
    glQuad[0].x = scaledX;
    glQuad[0].y = scaledY;
    glQuad[0].z = scaledZ;
    
    glQuad[1].x = -scaledX;
    glQuad[1].y = scaledY;
    glQuad[1].z = scaledZ;
    
    glQuad[2].x = -scaledX;
    glQuad[2].y = -scaledY;
    glQuad[2].z = scaledZ;
    
    glQuad[3].x = scaledX;
    glQuad[3].y = -scaledY;
    glQuad[3].z = scaledZ;
    
    // image coordinates for scale to fit
    if([[NSUserDefaults standardUserDefaults] boolForKey:@"Slideshow scale images"]){
        if(viewRatio > imgRatio){
            // image limited vertically - will not fill sides
            ul = NSMakePoint(0.0, (imgHeight*(1-imgRatio/viewRatio))/2.0);
            ur = NSMakePoint((float)imgWidth,(imgHeight*(1-imgRatio/viewRatio))/2.0);
            ll = NSMakePoint(0.0, imgHeight-(imgHeight*(1-imgRatio/viewRatio))/2.0);
            lr = NSMakePoint((float)imgWidth,imgHeight-(imgHeight*(1-imgRatio/viewRatio))/2.0);
        } else {
            // image limited horizontally - will not fill top/bottom
            ul = NSMakePoint((imgWidth*(1-viewRatio/imgRatio))/2.0, 0.0);
            ur = NSMakePoint(imgWidth-(imgWidth*(1-viewRatio/imgRatio))/2.0, 0.0);
            ll = NSMakePoint((imgWidth*(1-viewRatio/imgRatio))/2.0, (float)imgHeight);
            lr = NSMakePoint(imgWidth-(imgWidth*(1-viewRatio/imgRatio))/2.0, (float)imgHeight);
        }
    } else {
        // do not scale
        ul = NSMakePoint(0.0, 0.0);
        ur = NSMakePoint((float)imgWidth, 0.0);
        ll = NSMakePoint(0.0, (float)imgHeight);
        lr = NSMakePoint((float)imgWidth, (float)imgHeight);
    }
}

-(void)drawTexture
{
    if(!bmpRep){
        return;
    }

    if(![bmpRep bitmapData]){
        return;
    }
    
    if(ready){
        // Draw it!
        
        // draw black background rectangles covering field of view beside(!) image
        glColor4f(0.0f,0.0f,0.0f,myAlpha);
        
        if(viewRatio < imgRatio){
            //above
            glBegin(GL_QUADS);
                glVertex3f(backgroundQuad[0].x, backgroundQuad[0].y, backgroundQuad[0].z);// Upper Right
                glVertex3f(backgroundQuad[1].x, backgroundQuad[1].y, backgroundQuad[1].z);// Upper Left
                glVertex3f(glQuad[1].x, glQuad[1].y, glQuad[1].z);	// Lower Left
                glVertex3f(glQuad[0].x, glQuad[0].y, glQuad[0].z);	// Lower Right
            glEnd();
            // below
            glBegin(GL_QUADS);
                glVertex3f(glQuad[3].x, glQuad[3].y, glQuad[3].z);	// Upper Right
                glVertex3f(glQuad[2].x, glQuad[2].y, glQuad[2].z);	// Upper Left
                glVertex3f(backgroundQuad[2].x, backgroundQuad[2].y, backgroundQuad[2].z);// Lower Left
                glVertex3f(backgroundQuad[3].x, backgroundQuad[3].y, backgroundQuad[3].z);// Lower Right
            glEnd();
        } else {
        // left
            glBegin(GL_QUADS);
                glVertex3f(glQuad[1].x, glQuad[1].y, glQuad[1].z);	// Upper Right
                glVertex3f(backgroundQuad[1].x, backgroundQuad[1].y, backgroundQuad[1].z);// Upper Left
                glVertex3f(backgroundQuad[2].x, backgroundQuad[2].y, backgroundQuad[2].z);// Lower Left
                glVertex3f(glQuad[2].x, glQuad[2].y, glQuad[2].z);	// Lower Right
            glEnd();
            // right
            glBegin(GL_QUADS);
                glVertex3f(backgroundQuad[0].x, backgroundQuad[0].y, backgroundQuad[0].z);// Upper Right
                glVertex3f(glQuad[0].x, glQuad[0].y, glQuad[0].z);	// Upper Left
                glVertex3f(glQuad[3].x, glQuad[3].y, glQuad[3].z);	// Lower Left
                glVertex3f(backgroundQuad[3].x, backgroundQuad[3].y, backgroundQuad[3].z);// Lower Right
            glEnd();
        }
        
        // this is what I wanted to do, but it results in many "tearing" sorts of artifacts.
        /*
        glBegin(GL_QUADS);
            glVertex3f(backgroundQuad[0].x, backgroundQuad[0].y, backgroundQuad[0].z);	// Upper Right
            glVertex3f(backgroundQuad[1].x, backgroundQuad[1].y, backgroundQuad[1].z);	// Upper Left
            glVertex3f(backgroundQuad[2].x, backgroundQuad[2].y, backgroundQuad[2].z);	// Lower Left
            glVertex3f(backgroundQuad[3].x, backgroundQuad[3].y, backgroundQuad[3].z);	// Lower Right
	glEnd();
        */
        
        // draw texture
        glBindTexture(GL_TEXTURE_RECTANGLE_EXT, txID);
        glColor4f(myAlpha, myAlpha, myAlpha, myAlpha);

        glBegin(GL_QUADS);
            glTexCoord2f(ul.x, ul.y);				// Lower Left
            glVertex3f(glQuad[0].x, glQuad[0].y, glQuad[0].z);	// Upper Right
            
            glTexCoord2f(ur.x, ur.y);                           // Lower Right
            glVertex3f(glQuad[1].x, glQuad[1].y, glQuad[1].z);	// Upper Left
            
            glTexCoord2f(lr.x, lr.y);                           // Upper Right
            glVertex3f(glQuad[2].x, glQuad[2].y, glQuad[2].z);	// Lower Left
            
            glTexCoord2f(ll.x, ll.y);                           // Upper Left
            glVertex3f(glQuad[3].x, glQuad[3].y, glQuad[3].z);	// Lower Right
	glEnd();
        
    } else {
        NSLog(@"PSMGLImage not ready to draw");
    }
}

-(void)initGL
{
    GLuint err = 0;

    // Enable texturing
    glEnable(GL_TEXTURE_RECTANGLE_EXT);
    // Enable client storage
    glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 1);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, [bmpRep bytesPerRow]/([bmpRep bitsPerPixel] >> 3)); //imgWidth
    //[bmpRep bytesPerRow]/([bmpRep bitsPerPixel] >> 3)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    // Generate a texture ID
    glGenTextures(1, &txID);
    // Bind to our newly created texture ID
    glBindTexture(GL_TEXTURE_RECTANGLE_EXT, txID);
    // Specify the texture
    glTexImage2D(GL_TEXTURE_RECTANGLE_EXT, 0, bpp, imgWidth, imgHeight, 0, pixfmt, datatype, [bmpRep bitmapData]);
    err = glGetError();
    if(err)
    {
        NSLog(@"WARNING: OpenGL Error 0x%x while loading texture with glTexImage2D().", err);
        [self discardData];
        ready = NO;
        return;
    }
    [self genQuad];
    ready = YES;
}


-(void)loadTextureFromFile:(NSString*)filePath;
{
    // First, deallocate any image reps we might have.
    if(bmpRep){
        [bmpRep release];
        // We also need to delete our texture ID
        glDeleteTextures(1, &txID);
    }
    
    if(!filePath){
        return;
    }
    
    // Get raw bitmap data from the NSImage - old way
    // bmpRep = [[NSBitmapImageRep imageRepWithContentsOfFile:filePath] retain];
    // Load image and then get imagerep from tiff representation
    // this takes care of many pixel/bitdepth issues for some reason
    NSImage *sourceImage = [[NSImage alloc] initWithContentsOfFile:filePath];
    
    // check to see if the image is too big to be textured.  Scale if so...
    GLint max_texture_size = 0;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE,&max_texture_size);
    if(([[sourceImage bestRepresentationForDevice:nil] pixelsWide] > max_texture_size/2) || ([[sourceImage bestRepresentationForDevice:nil] pixelsHigh] > max_texture_size/2)){
        NSImage *newImage;
        
        // used to have EpegWrapper help with this...
        newImage = [[[NSImage alloc] initByReferencingFile:filePath] autorelease];
        [newImage setScalesWhenResized:YES];
        [newImage setSize:NSMakeSize(max_texture_size/2,max_texture_size/2)];
        
        // resize to max
        bmpRep = [[NSBitmapImageRep alloc] initWithData: [newImage TIFFRepresentation]];
    } else {
        bmpRep = [[NSBitmapImageRep alloc] initWithData: [sourceImage TIFFRepresentation]];
    }

    // Set the image height & width for convenience
    imgWidth = [bmpRep pixelsWide];
    imgHeight = [bmpRep pixelsHigh];
    imgRatio = (float)imgWidth/(float)imgHeight;
    [self genImageInfo:[filePath lastPathComponent]];
    [self getTextureInfo];
    [sourceImage release];
    // Run through our initialization routine
    [self initGL];
}

-(void)setTxID:(int)newTxID
{
    txID = newTxID;
}

-(void)discardData
{
    // Dump the bitmap
    if(bmpRep){
        [bmpRep release];
        bmpRep = nil;
    }

    // Dump the texture
    glDeleteTextures(1, &txID);

    // Clear the image width & height
    imgWidth = 0.0;
    imgHeight = 0.0;
    
    [self genImageInfo:@"No Info"];
}

-(void)dealloc
{
    [super dealloc];
}

-(void)getTextureInfo
{
    if([bmpRep hasAlpha]){
        bpp = GL_RGBA;
        pixfmt = GL_RGBA;
    } else {
        // need to be smart about image encoding - web images can be a mess
        switch([bmpRep samplesPerPixel]){
            case 1: // grayscale image
                bpp = GL_LUMINANCE;
                pixfmt = GL_LUMINANCE;
                break;
                
            case 2: // grayscale + alpha
                bpp = GL_LUMINANCE_ALPHA;
                pixfmt = GL_LUMINANCE;
                break;
                
            case 3:
                bpp = GL_RGB;
                pixfmt = GL_RGB;
                break;
                
            default:
                bpp = GL_RGB;
                pixfmt = GL_RGB;
                break;
        }
    }
    datatype = GL_UNSIGNED_BYTE;
}

-(void)genImageInfo:(NSString*)imgName
{
    NSMutableString *info = [NSMutableString stringWithFormat:@"Image Name:%@\n Size (w x h): %d x %d pixels ", imgName, imgWidth, imgHeight];
    
    if(!imageInfo){
        imageInfo = [[(GLString*)[GLString alloc] initWithString:info] retain];
    } else {
        [imageInfo setString:info];
    }
}

// transitions
-(void)setAlpha:(float)newAlpha
{
    myAlpha = newAlpha;
}

// cube placement methods
-(void)placeFront
{
    
}

-(void)placeLeft
{
    // x -> z, y -> y, z -> -x
    float newX = -glQuad[0].z;
    
    glQuad[0].z =  glQuad[0].x;
    glQuad[0].x = newX;
    
    glQuad[1].z =  glQuad[1].x;
    glQuad[1].x = newX;
    
    glQuad[2].z =  glQuad[2].x;
    glQuad[2].x = newX;
    
    glQuad[3].z =  glQuad[3].x;
    glQuad[3].x = newX;
    
    // background mask
    backgroundQuad[0].z =  backgroundQuad[0].x;
    backgroundQuad[0].x = newX;
    
    backgroundQuad[1].z =  backgroundQuad[1].x;
    backgroundQuad[1].x = newX;
    
    backgroundQuad[2].z =  backgroundQuad[2].x;
    backgroundQuad[2].x = newX;
    
    backgroundQuad[3].z =  backgroundQuad[3].x;
    backgroundQuad[3].x = newX;
}

-(void)placeRight
{
    // x -> -z, y -> y, z -> x
    float newX = glQuad[0].z;
    
    glQuad[0].z =  -glQuad[0].x;
    glQuad[0].x = newX;
    
    glQuad[1].z =  -glQuad[1].x;
    glQuad[1].x = newX;
    
    glQuad[2].z =  -glQuad[2].x;
    glQuad[2].x = newX;
    
    glQuad[3].z =  -glQuad[3].x;
    glQuad[3].x = newX;
    
    // background mask
    backgroundQuad[0].z =  -backgroundQuad[0].x;
    backgroundQuad[0].x = newX;
    
    backgroundQuad[1].z =  -backgroundQuad[1].x;
    backgroundQuad[1].x = newX;
    
    backgroundQuad[2].z =  -backgroundQuad[2].x;
    backgroundQuad[2].x = newX;
    
    backgroundQuad[3].z =  -backgroundQuad[3].x;
    backgroundQuad[3].x = newX;
}

-(void)placeBottom
{
    // x -> x, y -> -z, z -> y
    float newY = glQuad[0].z;
    
    glQuad[0].z = -glQuad[0].y;
    glQuad[0].y = newY;
    
    glQuad[1].z = -glQuad[1].y;
    glQuad[1].y = newY;
    
    glQuad[2].z = -glQuad[2].y;
    glQuad[2].y = newY;
    
    glQuad[3].z = -glQuad[3].y;
    glQuad[3].y = newY;
    
    // background mask
    backgroundQuad[0].z = -backgroundQuad[0].y;
    backgroundQuad[0].y = newY;
    
    backgroundQuad[1].z = -backgroundQuad[1].y;
    backgroundQuad[1].y = newY;
    
    backgroundQuad[2].z = -backgroundQuad[2].y;
    backgroundQuad[2].y = newY;
    
    backgroundQuad[3].z = -backgroundQuad[3].y;
    backgroundQuad[3].y = newY;
    
}

-(void)placeTop
{
    // x -> x, y -> z, z -> -y
    float newY = -glQuad[0].z;
    
    glQuad[0].z = glQuad[0].y;
    glQuad[0].y = newY;
    
    glQuad[1].z = glQuad[1].y;
    glQuad[1].y = newY;
    
    glQuad[2].z = glQuad[2].y;
    glQuad[2].y = newY;
    
    glQuad[3].z = glQuad[3].y;
    glQuad[3].y = newY;
    
    // background mask
    backgroundQuad[0].z = backgroundQuad[0].y;
    backgroundQuad[0].y = newY;
    
    backgroundQuad[1].z = backgroundQuad[1].y;
    backgroundQuad[1].y = newY;
    
    backgroundQuad[2].z = backgroundQuad[2].y;
    backgroundQuad[2].y = newY;
    
    backgroundQuad[3].z = backgroundQuad[3].y;
    backgroundQuad[3].y = newY;
}

-(void)placeFlip
{
    // x -> -x, y -> -y, z -> 0
    switch([[NSUserDefaults standardUserDefaults] integerForKey:@"Slideshow secondary transition"]){
        case 0:
        case 1: // vertical, flip y
            glQuad[0].x = glQuad[0].x;
            glQuad[0].y = -glQuad[0].y;
            glQuad[0].z = 0.1f;
            
            glQuad[1].x = glQuad[1].x;
            glQuad[1].y = -glQuad[1].y;
            glQuad[1].z = 0.1f;
            
            glQuad[2].x = glQuad[2].x;
            glQuad[2].y = -glQuad[2].y;
            glQuad[2].z = 0.1f;
            
            glQuad[3].x = glQuad[3].x;
            glQuad[3].y = -glQuad[3].y;
            glQuad[3].z = 0.1f;
            
            backgroundQuad[0].x = backgroundQuad[0].x;
            backgroundQuad[0].y = -backgroundQuad[0].y;
            backgroundQuad[0].z = 0.1f;
            
            backgroundQuad[1].x = backgroundQuad[1].x;
            backgroundQuad[1].y = -backgroundQuad[1].y;
            backgroundQuad[1].z = 0.1f;
            
            backgroundQuad[2].x = backgroundQuad[2].x;
            backgroundQuad[2].y = -backgroundQuad[2].y;
            backgroundQuad[2].z = 0.1f;
            
            backgroundQuad[3].x = backgroundQuad[3].x;
            backgroundQuad[3].y = -backgroundQuad[3].y;
            backgroundQuad[3].z = 0.1f;
            break;
            
        case 2:
        case 3: // horizontal, flip x
            glQuad[0].x = -glQuad[0].x;
            glQuad[0].y = glQuad[0].y;
            glQuad[0].z = 0.1f;
            
            glQuad[1].x = -glQuad[1].x;
            glQuad[1].y = glQuad[1].y;
            glQuad[1].z = 0.1f;
            
            glQuad[2].x = -glQuad[2].x;
            glQuad[2].y = glQuad[2].y;
            glQuad[2].z = 0.1f;
            
            glQuad[3].x = -glQuad[3].x;
            glQuad[3].y = glQuad[3].y;
            glQuad[3].z = 0.1f;
            
            backgroundQuad[0].x = -backgroundQuad[0].x;
            backgroundQuad[0].y = backgroundQuad[0].y;
            backgroundQuad[0].z = 0.1f;
            
            backgroundQuad[1].x = -backgroundQuad[1].x;
            backgroundQuad[1].y = backgroundQuad[1].y;
            backgroundQuad[1].z = 0.1f;
            
            backgroundQuad[2].x = -backgroundQuad[2].x;
            backgroundQuad[2].y = backgroundQuad[2].y;
            backgroundQuad[2].z = 0.1f;
            
            backgroundQuad[3].x = -backgroundQuad[3].x;
            backgroundQuad[3].y = backgroundQuad[3].y;
            backgroundQuad[3].z = 0.1f;
            break;
    }
}

@end
