//
//  ThumbnailMatrix.m
//  NetScrape
//
//  Created by John Pannell on Fri Sep 27 2002.
//  Copyright (c) 2001 Positive Spin Media. All rights reserved.
//

#import "ThumbnailMatrix.h"
#import "PSMActionImageCell.h"
#import "PSMWindowController.h"
#import "PSMAppController.h"
#import "PSMPandoraSession.h"
#import "PasteController.h"


@implementation ThumbnailMatrix
#pragma mark -
#pragma mark Creation/Destruction

- (id)initWithFrame:(NSRect)frameRect
{
    // default number of columns
    _numberOfColumns = [[NSUserDefaults standardUserDefaults] integerForKey:@"number of columns"];
    _numImages = 0;
    drawForDrop = NO;
    [self registerForDraggedTypes:[NSArray arrayWithObjects:NSFilenamesPboardType, nil]];
    
    // nothing placed yet
    _currentRow = 0;
    _currentColumn = 0;
    
    // create the cell prototype and style it
    id cell = [[PSMActionImageCell alloc] init];
    [cell setImageAlignment:NSImageAlignCenter];
    [cell setImageFrameStyle:NSImageFrameNone];
    [cell setImageScaling:NSScaleProportionally];
    [cell setEnabled:NO];

    // create the actual matrix of buttons
    [super initWithFrame:frameRect
                    mode:NSListModeMatrix
               prototype:cell
            numberOfRows:0
         numberOfColumns:_numberOfColumns];

    // put a white background behind cells
    [self setDrawsBackground:YES];
    [self setBackgroundColor:[NSColor whiteColor]];

    // configure the size and spacing of buttons
    // use the width of the superview (the scroller)
    
    [self setCellSize:NSMakeSize(100,100)];
    [self setIntercellSpacing:NSMakeSize(0,0)];
    [self setSelectionByRect:NO];
    
    [cell release];

    return self;
}

#pragma mark -
#pragma mark Accessors

- (NSEvent *)lastMouseDownEvent
{
    return _lastMouseDownEvent;
}

- (void)setLastMouseDownEvent:(NSEvent *)event
{
    [event retain];
    [_lastMouseDownEvent release];
    _lastMouseDownEvent = event;
}

- (BOOL)drawForDrop
{ 
    return drawForDrop;
}

- (id)engine
{
    return _engine;
}

- (void)setEngine:(id)engine
{
    [engine retain];
    [_engine release];
    _engine = engine;
}

- (NSPoint)dropMouseLocation
{
    return dropMouseLocation;
}

- (void)setNumberOfColumns:(int)number
{
    // resize the matrix to a different number of columns
    // determine how many cells I have
    int i, rowCount, cellCount;
    cellCount = (_currentRow * _numberOfColumns + _currentColumn);
    NSArray *temp = [self selectedCells];
    
    // determine number of new rows with new columns
    rowCount = cellCount/number + 1;
    
    // set new numberOfColumns for the record
    _numberOfColumns = number;
    [[NSUserDefaults standardUserDefaults] setInteger:number forKey:@"number of columns"];
    
    // determine new placement
    _currentRow = rowCount - 1;
    _currentColumn = cellCount%number;
    
    // resize
    [self renewRows:rowCount columns:number];
    
    // reselect the selected cells
    for(i = 0; i < cellCount; i++){
        if([temp containsObject:[[self cells] objectAtIndex:i]]){
            [self setSelectionFrom:i to:i anchor:i highlight:YES];
        }
    }
}

- (int)numberOfColumns
{
    return _numberOfColumns;
}

- (int)numImages
{
    return _numImages;
}

- (unsigned long long)diskSpace
{
    // calculate disk space occupied by objects represented by cells
    NSEnumerator *e = [[self cells] objectEnumerator];
    unsigned long long diskSpace = 0;
    PSMActionImageCell *thisCell;
    while(thisCell = [e nextObject]){
        if([thisCell representedObject]){
            diskSpace += [(PSMWebFile *)[thisCell representedObject] actualContentLength];
        }
    }
    return diskSpace;
}

- (NSString *)imageCountString
{
    // return a string with the proper number of images
    int diskSpace = [self diskSpace];
    NSString *size, *returnString;
    
    if((diskSpace/1024) < 1000){
        // measure in K
        int thisSize = floor(diskSpace/1024);
        size = [NSString stringWithFormat:@"%d K",thisSize];
    } else {
        // measure in M
        float thisSize = diskSpace/(1024.0*1024.0);
        size = [NSString stringWithFormat:@"%0.2f M",thisSize];
    }
    
    if(_numImages != 1){
        returnString = [NSString stringWithFormat:@"%d images, %@", _numImages, size];
        return returnString;
    } else {
        returnString = [NSString stringWithFormat:@"%d image, %@", _numImages, size];
        return returnString;
    }
}


#pragma mark -
#pragma mark Functionality

- (void)addScrapedFile:(id)object
{
    // place an image into the matrix
    // _currentRow and _currentColumn contain the position this should be placed
    // allocate new row if needed
    if(_currentRow > [self numberOfRows] - 1){
        [self addRow];
        [self sizeToCells];
    }
    
    // place image
    if([self cellAtRow:_currentRow column:_currentColumn]){
        [[self cellAtRow:_currentRow column:_currentColumn] setEnabled:YES];
        [[self cellAtRow:_currentRow column:_currentColumn] setImage:[object thumbnail]];
        [[self cellAtRow:_currentRow column:_currentColumn] setRepresentedObject:object];
    }
    _numImages++;
    [self setNeedsDisplay:YES];
    
    // handle if this is the first image (0 for row and column and this view is showing)
    if((_currentRow == 0) && (_currentColumn == 0) && ([[[self window] firstResponder] isEqual:self])){
        [self selectCellAtRow:_currentRow column:_currentColumn];
        [[[self window] delegate] userClick:self];
    }
    
    // get ready for the next one
    _currentColumn++;
    if(_currentColumn >= _numberOfColumns){
        _currentColumn = 0;
        _currentRow++;
    }
    
    // update the footer if I'm showing
    if([[[[[[[self window] delegate] previewTabView] selectedTabViewItem] view] documentView] isEqual:self])
        [[[self window] delegate] updatePreviewFooter];
}

- (void)injectPastedFile:(PSMWebFile *)file atIndex:(int)index
{
    [file retain];
    // allocate new row if needed
    if(_currentRow > [self numberOfRows] - 1){
        [self addRow];
        [self sizeToCells];
    }
    
    // get ready for the next one
    _currentColumn++;
    if(_currentColumn >= _numberOfColumns){
        _currentColumn = 0;
        _currentRow++;
    }
    _numImages++;
    
    NSMutableArray *tempArray = [[NSMutableArray alloc] initWithArray:[self cells]];
    // make a new cell
    id cell = [[self prototype] copy];
    [cell setEnabled:YES];
    [cell setRepresentedObject:file];
    [cell setImage:[file thumbnail]];
    
    // place cell
    [tempArray insertObject:cell atIndex:index];
    [cell release];
    
    // run through array and reassign cells
    [self reassignCellsFromArray:tempArray];
    
    // select the new cell
    [self deselectAllCells];
    [self setSelectionFrom:index to:index anchor:index highlight:YES];
    [tempArray release];
    [file release];
    
    [self setNeedsDisplay:YES];
    if([[[self window] firstResponder] isEqual:self]){
        [[[self window] delegate] userClick:self];
        [[[self window] delegate] updatePreviewFooter];
    }
}

- (void)moveSelectedCellsToIndex:(int)injectionIndex
{
    NSMutableArray *tempArray = [[NSMutableArray alloc] initWithArray:[self cells]];
    // pull selected cells and replace with blanks
    NSEnumerator *sc = [[self selectedCells] objectEnumerator];
    PSMActionImageCell *thisCell;
    while(thisCell = [sc nextObject]){
        [tempArray replaceObjectAtIndex:[tempArray indexOfObject:thisCell] withObject:[[self prototype] copy]];
    }
    
    // find the index of the marker slide
    int startPosition = injectionIndex;
    NSEnumerator *e = [[self selectedCells] reverseObjectEnumerator];
    while(thisCell = [e nextObject]){
        [tempArray insertObject:thisCell atIndex:startPosition];
    }
    
    // count dead cells to be removed before startPosition
    int i, end = startPosition;
    for(i = 0; i < end; i++){
        if([[tempArray objectAtIndex:i] image] == nil){
            startPosition--;
        }
    }
    
    // sort dead cells to end
    NSEnumerator *te = [tempArray reverseObjectEnumerator];
    BOOL nonBlankCell = NO;
    while(thisCell = [te nextObject]){
        if([thisCell image] == nil){
            if(nonBlankCell){
                // must pull blank cells from the mix and move to end
                [thisCell retain];
                [tempArray removeObject:thisCell];
                [tempArray addObject:thisCell];
                [thisCell release];
            }
        } else {
            nonBlankCell = YES;
        }
    }
    
    int endPosition = startPosition + [[self selectedCells] count] - 1;
    // run through array and reassign cells
    [self reassignCellsFromArray:tempArray];
    
    // select the originally selected cells
    [self setSelectionFrom:startPosition to:endPosition anchor:startPosition highlight:YES];
    [tempArray release];
}

- (NSMutableArray *)slideShowPaths
{
    // list of file paths for the slideshow
    NSMutableArray *returnArray = [[NSMutableArray alloc] init];
    PSMActionImageCell *cell;
    NSEnumerator *e = [[self cells] objectEnumerator];
    while(cell = [e nextObject]){
        if([cell representedObject]){ // some cells are blank :-)
            if([[NSBitmapImageRep imageUnfilteredFileTypes] containsObject:[[[cell representedObject] storagePath] pathExtension]]){
                [returnArray addObject:[[cell representedObject] storagePath]];
            }
        }
    }
    return returnArray;
}

#pragma mark -
#pragma mark Responder

-(BOOL)acceptsFirstResponder
{
    // I'll take it if I'm visible
    if([[[self window] delegate] isToolboxCollapsed]){
        return NO;
    } else {
        return YES;
    }
}

- (BOOL)becomeFirstResponder
{
    [self setNextKeyView:[[[self window] delegate] URLField]];
    [self setNeedsDisplay:YES];
    return YES;
}

- (BOOL)resignFirstResponder
{
    [self setNeedsDisplay:YES];
    return YES;
}

- (void)keyDown:(NSEvent *)theEvent 
{ 
    // nothing but arrow keys and delete
    if ([theEvent type] == NSKeyDown){
        switch ([[theEvent characters] characterAtIndex:0]) 
        { 
            case NSUpArrowFunctionKey:
                [self moveUp:self];
                break;
            case NSDownArrowFunctionKey:
                [self moveDown:self];
                break;
            case NSLeftArrowFunctionKey: 
                [self moveLeft:self];
                break;
            case NSRightArrowFunctionKey: 
                [self moveRight:self];
                break;
            case NSTabCharacter:
                [[self window] makeFirstResponder:[self nextKeyView]];
                break;
            case NSDeleteCharacter:
                //[self deleteSelectedImage];
                break;
            default:
                // make the text field key and send event
                [[self window] makeFirstResponder:[self nextKeyView]];
                [[self window] sendEvent:theEvent];
                break;
        } 
    } else {
        [[self window] sendEvent:theEvent];
    }
}

-(void)moveRight:(id)sender
{
    if([self selectedColumn] < ([self numberOfColumns] - 1)){
        // can't pick empty cell
        if([[self cellAtRow:[self selectedRow] column:[self selectedColumn]+1] image] != nil){
            // mark as dirty
            [self setNeedsDisplay:YES];
            // highlight next cell
            [self selectCellAtRow:[self selectedRow] column:[self selectedColumn]+1];
            [[[self window] delegate] userClick:self];
        }
    } else {
        if([self selectedRow] < [self numberOfRows]){
            // can't pick empty cell
            if([[self cellAtRow:[self selectedRow]+1 column:0] image] != nil){
                // mark as dirty
                [self setNeedsDisplay:YES];
                // highlight next cell
                [self selectCellAtRow:[self selectedRow]+1 column:0];
                [[[self window] delegate] userClick:self];
            }
        }
    }
    [self scrollRectToVisible:[self cellFrameAtRow:[self selectedRow] column:[self selectedColumn]]];
}
-(void)moveLeft:(id)sender
{
    if([self selectedColumn] > 0){
        // mark as dirty
        [self setNeedsDisplay:YES];
        // highlight next cell
        [self selectCellAtRow:[self selectedRow] column:[self selectedColumn]-1];
        [[[self window] delegate] userClick:self];
    } else {
        // mark as dirty
        [self setNeedsDisplay:YES];
        // highlight next cell
        if([self selectedRow] > 0){
            [self selectCellAtRow:[self selectedRow]-1 column:([self numberOfColumns] - 1)];
            [[[self window] delegate] userClick:self];
        }
    }
    [self scrollRectToVisible:[self cellFrameAtRow:[self selectedRow] column:[self selectedColumn]]];
}
-(void)moveUp:(id)sender
{
    if([self selectedRow] > 0){
        // mark as dirty
        [self setNeedsDisplay:YES];
        // highlight next cell
        [self selectCellAtRow:[self selectedRow]-1 column:[self selectedColumn]];
        [[[self window] delegate] userClick:self];
    }
    [self scrollRectToVisible:[self cellFrameAtRow:[self selectedRow] column:[self selectedColumn]]];
}
-(void)moveDown:(id)sender
{
    if([self selectedRow] < [self numberOfRows]){
        // can't pick empty cell
        if([[self cellAtRow:[self selectedRow]+1 column:[self selectedColumn]] image] != nil){
            // mark as dirty
            [self setNeedsDisplay:YES];
            // highlight next cell
            [self selectCellAtRow:[self selectedRow]+1 column:[self selectedColumn]];
            [[[self window] delegate] userClick:self];
        }
    }
    [self scrollRectToVisible:[self cellFrameAtRow:[self selectedRow] column:[self selectedColumn]]];
}



#pragma mark -
#pragma mark Mouse Tracking
 
- (void)mouseDown:(NSEvent *)theEvent
{
    // take focus back if needed
    if([[self window] firstResponder] != self){
        [[self window] makeFirstResponder:self];
    }
    // can't pick a cell with no image
    int row, column, flags = [theEvent modifierFlags];
    [self setLastMouseDownEvent:theEvent];
    if([self getRow:&row column:&column forPoint:[self convertPoint:[theEvent locationInWindow] fromView:[[self window] contentView]]]){
        if([[self cellAtRow:row column:column] image] == nil){
            [self setLastMouseDownEvent: nil];
            // ignore this event
        } else {
            // override super implementation so dragging events still come through
            // look for Command- or Shift-click
            if(flags & NSCommandKeyMask){
                // discontinuous selection - if selected, unselect, otherwise select
                if([[self selectedCells] containsObject:[self cellAtRow:row column:column]]){
                    // unselect this one
                    NSMutableArray *temp = [[self selectedCells] mutableCopy];
                    [temp removeObject:[self cellAtRow:row column:column]];
                    [self deselectAllCells];
                    NSEnumerator *e = [temp objectEnumerator];
                    PSMActionImageCell *thisCell;
                    while(thisCell = [e nextObject]){
                        int row, col;
                        if([self getRow:&row column:&col ofCell:thisCell]){
                            int thisPos = (row * _numberOfColumns + col);
                            [self setSelectionFrom:thisPos to:thisPos anchor:thisPos highlight:YES];
                        }
                    }
                } else {
                    // select this one
                    int thisPos = (row * _numberOfColumns + column);
                    [self setSelectionFrom:thisPos to:thisPos anchor:thisPos highlight:YES];
                }
            }
                
            
            if(flags & NSShiftKeyMask){
                // continuous selection
                int prevRow, prevCol;
                if([self getRow:&prevRow column:&prevCol ofCell:[self selectedCell]]){
                    int startPos = (prevRow * _numberOfColumns + prevCol);
                    int endPos = (row * _numberOfColumns + column);
                    [self setSelectionFrom:startPos to:endPos anchor:startPos highlight:YES];
                }
            }
            
            if(!((flags & NSShiftKeyMask) || (flags & NSCommandKeyMask))){
                // if selected already, do nothing
                if(![[self selectedCells] containsObject:[self cellAtRow:row column:column]]){
                    [self selectCellAtRow:row column:column];
                }
            }
            
            [self sendAction];
        }
    } else {
        [super mouseDown:theEvent];
    }
    // clean up highlighting
    [self setNeedsDisplay:YES];
}

- (void)mouseDragged:(NSEvent *)theEvent
{
    NSPasteboard *pboard = [NSPasteboard pasteboardWithName:NSDragPboard];
    if([self lastMouseDownEvent] != nil){
        NSPoint trackingStartPoint = [self convertPoint:[[self lastMouseDownEvent] locationInWindow] fromView:nil];
        NSPoint currentPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];
        float dx = fabs(currentPoint.x - trackingStartPoint.x);
        float dy = fabs(currentPoint.y - trackingStartPoint.y);
        float distance = sqrt(dx * dx + dy * dy);
        if(distance >= 10){
            NSPoint imagePoint = [self convertPoint:[[self lastMouseDownEvent] locationInWindow] fromView:nil];
            NSImage *sourceImage, *translucent;
            NSPoint cellOrigin;
            // get image under mouse
            int row, col;
            if([self getRow:&row column:&col forPoint:imagePoint]){
                sourceImage = [[[self cellAtRow:row column:col] representedObject] thumbnail];
                cellOrigin =  [self cellFrameAtRow:row column:col].origin;
            } else {
                sourceImage = [[[[self selectedCells] objectAtIndex:0] representedObject] thumbnail];
                cellOrigin =  [self cellFrameAtRow: [self selectedRow] column: [self selectedColumn]].origin;
            }
            
            // center the slide on the mouse cursor
            int height = [sourceImage size].height;
            int width = [sourceImage size].width; 
            
            // good looking translucent image
            translucent = [[[NSImage alloc] initWithSize: [sourceImage size]] autorelease];
            [translucent lockFocus];
            [sourceImage dissolveToPoint: NSZeroPoint fraction: 0.7];
            int numCells = [[self selectedCells] count];
            if(numCells > 1){
                // badge with number of images
                NSImage *badgeImage = nil;
                if(numCells < 100){
                    badgeImage = [NSImage imageNamed:@"dragBadge1-2"];
                }
                if((numCells >= 100) && (numCells < 1000)){
                    badgeImage = [NSImage imageNamed:@"dragBadge3"];
                }
                if((numCells >= 1000) && (numCells < 10000)){
                    badgeImage = [NSImage imageNamed:@"dragBadge4"];
                }
                if((numCells >= 10000) && (numCells < 100000)){
                    badgeImage = [NSImage imageNamed:@"dragBadge5"];
                }
                // compose onto image
                [badgeImage compositeToPoint:NSMakePoint([translucent size].width-[badgeImage size].width, 0) operation:NSCompositeSourceOver];
                NSAttributedString* stringToDraw = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%d",[[self selectedCells] count]] attributes:[NSDictionary dictionaryWithObjectsAndKeys: [NSColor whiteColor], NSForegroundColorAttributeName, [NSFont fontWithName:@"Helvetica-Bold" size:11], NSFontAttributeName, nil]];
                NSSize stringSize = [stringToDraw size];
                NSPoint destPoint = NSMakePoint((([translucent size].width-[badgeImage size].width)+([badgeImage size].width - stringSize.width) / 2) + 0.5, (([badgeImage size].height - stringSize.height) / 2)+0.5);
                [stringToDraw drawAtPoint:destPoint];
            }
            [translucent unlockFocus];
            
            [pboard declareTypes:[NSArray arrayWithObject:NSFilenamesPboardType] owner:self];
            [pboard setPropertyList:[self selectedFiles] forType:NSFilenamesPboardType];
            
            NSPoint dragPoint = [self convertPoint:[[self window] mouseLocationOutsideOfEventStream] fromView:[[self window] contentView]];
            dragPoint.x -= width * (dragPoint.x - cellOrigin.x) / width;
            dragPoint.y += height * (cellOrigin.y + height - dragPoint.y) / height;
            [self dragImage:translucent at:dragPoint offset:[ThumbnailMatrix subtractPoints:[[self window] mouseLocationOutsideOfEventStream] from:[[self lastMouseDownEvent] locationInWindow]] event:[self lastMouseDownEvent] pasteboard:pboard source:self slideBack:YES];
        }
    }
}

#pragma mark -
#pragma mark Cut, Copy, Paste

- (IBAction)cut:(id)sender
{
    if([[self selectedCell] representedObject] != nil){
        // remove slides from preview and place info on pasteboard
        NSPasteboard *pb = [NSPasteboard generalPasteboard];
        [self writeSlideToPasteboard:pb];
        // remove from view
        [self deleteSelectedCellsWithoutTrashing];
        [[[self window] delegate] updatePreviewFooter];
    }
}

- (IBAction)copy:(id)sender
{
    if([[self selectedCell] representedObject] != nil){
        NSPasteboard *pb = [NSPasteboard generalPasteboard];
        [self writeSlideToPasteboard:pb];
    }
}

- (IBAction)paste:(id)sender
{
    NSPasteboard *pb = [NSPasteboard generalPasteboard];
    if(![self readSlideFromPasteboard:pb]){
        NSBeep();
    }
}


- (IBAction)clear:(id)sender
{
    if([[self selectedCell] representedObject] != nil){
        if([[[self window] delegate] canClearSlides:self])
            [self deleteSelectedCellsWithTrashing];
        if([[[self window] firstResponder] isEqual:self])
            [[[self window] delegate] updatePreviewFooter];
    }
}

- (BOOL)validateToolbarItem:(NSToolbarItem *)theItem
{
    if([[theItem itemIdentifier] isEqualToString:@"TrashItem"]){
        // must have something selected
        if([[self selectedCell] representedObject] != nil){
            return YES;
        } else {
            return NO;
        }
    }
    
    return NO;
}

- (BOOL)validateMenuItem:(NSMenuItem*)anItem
{
    if ([anItem action] == @selector(paste:)){
        if([[[NSPasteboard generalPasteboard] types] containsObject:NSFilenamesPboardType]){
            return YES;
        }
        return NO;
    }
    
    if ([anItem action] == @selector(cut:)){
        if([[self selectedCell] representedObject]){
            return YES;
        }
        return NO;
    }
    
    if ([anItem action] == @selector(copy:)){
        if([[self selectedCell] representedObject]){
            return YES;
        }
        return NO;
    }
    
    if ([anItem action] == @selector(clear:)){
        if([[self selectedCell] representedObject]){
            return YES;
        }
        return NO;
    }
    
    if ([anItem action] == @selector(selectAll:)){
        return YES;
    }
    
    return NO;
}

- (void)writeSlideToPasteboard:(NSPasteboard *)pb
{
    // write slide data (image file path) to pasteboard
    NSArray *types = [NSArray arrayWithObject:NSFilenamesPboardType];
    [pb declareTypes:types owner:self];
    
    if(![pb setPropertyList:[self selectedFiles] forType:NSFilenamesPboardType]){
        NSBeep();
    }
}

- (BOOL)readSlideFromPasteboard:(NSPasteboard *)pb
{
    // is there anything to read?
    if(![[pb types] containsObject:NSFilenamesPboardType]){
        return NO;
    }
    
    // create a pasteController to paste in files
    PasteController *pc = [[PasteController alloc] initWithFiles:[pb propertyListForType:NSFilenamesPboardType] matrix:self injectionIndex:[self numImages]];
    [pc startPaste];
    
    return YES;
}

- (void)deleteSelectedCellsWithoutTrashing
{
    // remove cells, repopulate matrix from temp array
    NSMutableArray *tempArray = [[NSMutableArray alloc] initWithArray:[self cells]];
    [tempArray removeObjectsInArray:[self selectedCells]];
    PSMActionImageCell *deleteCell;
    NSEnumerator *de = [[self selectedCells] objectEnumerator];
    int i, lastRow, lastColumn, lastImageNum = 0, numDeletedCells = [[self selectedCells] count], numToRemove;
    
    _numImages = _numImages - numDeletedCells;
    
    // find out the last selected cell so I can pick it's neighbor after deleting
    while(deleteCell = [de nextObject]){
        // don't leak the scrapedFile objects
        [[[self engine] scrapedFiles] removeObject:[deleteCell representedObject]]; // remove from model
        int currentCount = [[self engine] objectCount];
        currentCount--;
        [[[self engine] objectController] setValue:[NSNumber numberWithInt:currentCount] forKeyPath:@"selection.objectCount"]; // decrement object count
        //[[deleteCell representedObject] release];
        
        [self getRow:&lastRow column:&lastColumn ofCell:deleteCell];
        int thisImageNum = _numberOfColumns * lastRow + lastColumn - numDeletedCells + 1;
        if(thisImageNum > lastImageNum){
            lastImageNum = thisImageNum;
        }
        if(lastImageNum > (_numImages - 1)){
            lastImageNum = (_numImages - 1);
        }
    }
    
    // remove rows if necessary
    int neededRows = _numImages/_numberOfColumns;
    if((_numImages % _numberOfColumns) != 0){
        // add one because there was a remainder
        neededRows++;
    }
    numToRemove = [self numberOfRows] - neededRows;
    for(i=0;i<numToRemove;i++){
        [self removeRow:([self numberOfRows]-1)];
    }
    
    // run through array and reassign cells
    [self reassignCellsFromArray:tempArray];
    
    // clean up
    _currentRow = _numImages/_numberOfColumns;
    _currentColumn = _numImages%_numberOfColumns;
    
    // fill rest of row with blank cells
    for(i = _currentColumn; i < _numberOfColumns; i++){
        [self putCell:[[self prototype] copy] atRow:_currentRow column:i];
    }
    if(_currentColumn >= _numberOfColumns){
        _currentColumn = 0;
        _currentRow++;
    }
    // select the correct image
    lastRow = lastImageNum/_numberOfColumns;
    lastColumn = lastImageNum%_numberOfColumns;
    [self selectCellAtRow:lastRow column:lastColumn];
    
    [self sizeToCells];
    [self setNeedsDisplay:YES];
    [[[self window] delegate] userClick:self];
    [tempArray release];
}


- (void)deleteSelectedCellsWithTrashing
{
    // indentify the source files
    NSMutableSet *sourceFiles = [NSMutableSet set];
    NSEnumerator *sources = [[self selectedCells] objectEnumerator];
    PSMActionImageCell *thisCell;
    while(thisCell = [sources nextObject]){
        [sourceFiles addObject:[[thisCell representedObject] storagePath]]; // record file paths for later
    }
    
    // identify the first non-selected cell after the first selected cell
    // for selection after removal is done
    BOOL withinSelection = NO;
    int finalCellNumberToHighlight = 0;
    NSEnumerator *cellPicker = [[self cells] objectEnumerator];
    while(thisCell = [cellPicker nextObject]){
        //NSLog(@"%d, %d, %d", withinSelection, [thisCell isHighlighted], finalCellNumberToHighlight);
        if(!withinSelection)
            finalCellNumberToHighlight++;
        if(!withinSelection && [thisCell isHighlighted])
            withinSelection = YES;
        if(withinSelection && ![thisCell isHighlighted]){
            finalCellNumberToHighlight--;
            break; // got the number
        }
    }
    
    // pull cells from matrices
    NSEnumerator *sessionsEnumerator = [[[NSApp delegate] sessionArray] objectEnumerator];
    ThumbnailMatrix *thisMatrix;
    while(thisMatrix = [[sessionsEnumerator nextObject] thumbnailMatrix]){
        [thisMatrix removeCellsWithFiles:sourceFiles];
    }
    
    // delete from disk and pasteboard
    NSPasteboard *pb = [NSPasteboard generalPasteboard];
    NSEnumerator *files = [sourceFiles objectEnumerator];
    NSString *thisPath;
    while(thisPath = [files nextObject]){
        // disk
        int tag = 0;
        if([[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation source:[thisPath stringByDeletingLastPathComponent] destination:@"" files:[NSArray arrayWithObject:[thisPath lastPathComponent]] tag:&tag]){
            [[NSWorkspace sharedWorkspace] noteFileSystemChanged];
        }
        
        // pasteboard
        if([[pb types] containsObject:NSFilenamesPboardType]){
            NSMutableArray *tempReplacement = [[pb propertyListForType:NSFilenamesPboardType] mutableCopy];
            if([tempReplacement containsObject:thisPath]){
                [tempReplacement removeObject:thisPath];
                if([tempReplacement count] > 0){
                    [pb setPropertyList:tempReplacement forType:NSFilenamesPboardType];
                } else {
                    // nothing left
                    [pb declareTypes:nil owner:nil];
                }
            }
        }
    }
    
    // make sound
    /*
    NSSound *trashSound = [[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/Dock/Drag To Trash.aif" byReference:YES];
    if(trashSound){
        [trashSound play];
    }
     */
    
    // position the selection to the appropriate spot
    int lastRow, lastColumn;
    //NSLog(@"%d", finalCellNumberToHighlight);
    lastRow = finalCellNumberToHighlight/_numberOfColumns;
    lastColumn = finalCellNumberToHighlight%_numberOfColumns;
    [self selectCellAtRow:lastRow column:lastColumn];
    
    // highlight new image
    [[[self window] delegate] userClick:self];
}

- (void)removeCellsWithFiles:(NSMutableSet *)sourceFiles
{
    // must go through my cells and delete ones with file in given set
    // remove cells, repopulate matrix from temp array
    NSMutableArray *tempArray = [[NSMutableArray alloc] initWithArray:[self cells]];
    PSMActionImageCell *thisCell, *referenceCell = nil;
    NSEnumerator *de = [tempArray objectEnumerator];
    int i, lastRow, lastColumn, lastImageNum = 0, numToRemove;
    
    // find out the last selected cell so I can pick it's neighbor after deleting
    while(thisCell = [de nextObject]){
        if([sourceFiles containsObject:[[thisCell representedObject] storagePath]]){
            [[[self engine] scrapedFiles] removeObject:[thisCell representedObject]]; // remove from model
            int currentCount = [[self engine] objectCount];
            currentCount--;
            [[[self engine] objectController] setValue:[NSNumber numberWithInt:currentCount] forKeyPath:@"selection.objectCount"]; // decrement object count
            //[[thisCell representedObject] release];
            _numImages = _numImages - 1;
            
            [self getRow:&lastRow column:&lastColumn ofCell:thisCell];
            int thisImageNum = _numberOfColumns * lastRow + lastColumn;
            if(thisImageNum > lastImageNum){
                lastImageNum = thisImageNum;
            }
            if(lastImageNum > (_numImages - 1)){
                lastImageNum = (_numImages - 1);
            }
            
            [tempArray replaceObjectAtIndex:[tempArray indexOfObject:thisCell] withObject:[[self prototype] copy]];
        } else {
            // NEW - save selection for later - attempt to keep selection correctly after delete
            if([[self selectedCells] containsObject:thisCell])
                referenceCell = thisCell;
        }
    }
    
    // remove rows if necessary
    int neededRows = _numImages/_numberOfColumns;
    if((_numImages % _numberOfColumns) != 0){
        // add one because there was a remainder
        neededRows++;
    }
    numToRemove = [self numberOfRows] - neededRows;
    for(i=0;i<numToRemove;i++){
        [self removeRow:([self numberOfRows]-1)];
    }
    
    // sort dead cells to end
    NSEnumerator *te = [tempArray reverseObjectEnumerator];
    BOOL nonBlankCell = NO;
    while(thisCell = [te nextObject]){
        if([thisCell image] == nil){
            if(nonBlankCell){
                // must pull blank cells from the mix and move to end
                [thisCell retain];
                [tempArray removeObject:thisCell];
                [tempArray addObject:thisCell];
                [thisCell release];
            }
        } else {
            nonBlankCell = YES;
        }
    }
    
    // run through array and reassign cells
    [self reassignCellsFromArray:tempArray];
    
    // clean up
    _currentRow = _numImages/_numberOfColumns;
    _currentColumn = _numImages%_numberOfColumns;
    
    // fill rest of row with blank cells
    for(i = _currentColumn; i < _numberOfColumns; i++){
        [self putCell:[[self prototype] copy] atRow:_currentRow column:i];
    }
    if(_currentColumn >= _numberOfColumns){
        _currentColumn = 0;
        _currentRow++;
    }
    // select the correct image
    lastRow = lastImageNum/_numberOfColumns;
    lastColumn = lastImageNum%_numberOfColumns;
    [self selectCellAtRow:lastRow column:lastColumn];
    
    // NEW - attempt to set selection based on saved reference cell
    int s_row = 0, s_col = 0;
    if(referenceCell != nil){
        [self getRow:&s_row column:&s_col ofCell:referenceCell];
        [self selectCellAtRow:s_row column:s_col];
    }
    
    [self sizeToCells];
    [self setNeedsDisplay:YES];
    if(self == [[self window] firstResponder]){
        //[(ScrapeController *)[NSApp delegate] userClick:self];
    }
    [tempArray release];
}

#pragma mark -
#pragma mark Drag And Drop

// drag and drop
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
    return YES;
}

- (unsigned int)draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
    if(isLocal){
        // drop on myself
        return NSDragOperationMove;
    } else {
        // drop on other app
        if([[NSApp delegate] canDropToFinder]){
            return NSDragOperationCopy | NSDragOperationDelete;
        } else {
            return NSDragOperationDelete;
        }
    }
}

- (void)draggedImage:(NSImage *)anImage endedAt:(NSPoint)aPoint operation:(NSDragOperation)operation
{
    if(operation == NSDragOperationDelete){
        // move the selected files to trash
        [self deleteSelectedCellsWithTrashing];
        /*
        NSSound *trashSound = [[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/Dock/Drag To Trash.aif" byReference:YES];
        if(trashSound){
            [trashSound play];
        }
         */
    }
    
    /*
    if(operation == NSDragOperationCopy){
        NSSound *pasteSound = [[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/system/Volume Mount.aif" byReference:YES];
        if(pasteSound){
            [pasteSound play];
        }
    }
     */
    
    // set restriction for non-registered app
    NSPoint myScreenOrigin = [[self window] convertBaseToScreen:[self convertPoint:[self frame].origin toView:nil]];
    NSRect myScreenRect = NSMakeRect(myScreenOrigin.x, myScreenOrigin.y, [self frame].size.width, [self frame].size.height);
    if((![[NSApp delegate] isRegistered]) && !(NSPointInRect(aPoint,myScreenRect))){
        [[NSApp delegate] setCanDropToFinder:NO];
        // run warning panel
        NSAlert *noSaveWarning = [NSAlert alertWithMessageText:@"Demonstration Version" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:@"The demo version of Pandora restricts the saving of collected files, and allows only one drop to another application per session.  Drag and drop to other applications will not work for the rest of this session. Please register Pandora to unlock all of it's powerful features!"];
        [noSaveWarning beginSheetModalForWindow:[self window] modalDelegate:nil didEndSelector:nil contextInfo:nil];
    }
}

- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent
{
    return YES;
}

// and drop
- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
    if ([sender draggingSource] == self) {
        drawForDrop = YES;
        dropMouseLocation = [sender draggingLocation];
        [self setNeedsDisplay];
        return NSDragOperationMove;
    } else {
        if ([[[sender draggingPasteboard] types] indexOfObject:NSFilenamesPboardType] != NSNotFound) {
            drawForDrop = YES;
            dropMouseLocation = [sender draggingLocation];
            [self setNeedsDisplay];
            return NSDragOperationCopy;
        } else {
            return NSDragOperationNone;
        }
    }
}

- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
{
    if ([sender draggingSource] == self) {
        drawForDrop = YES;
        dropMouseLocation = [sender draggingLocation];
        [self setNeedsDisplay];
        return NSDragOperationMove;
    } else {
        if ([[[sender draggingPasteboard] types] indexOfObject:NSFilenamesPboardType] != NSNotFound) {
            drawForDrop = YES;
            dropMouseLocation = [sender draggingLocation];
            [self setNeedsDisplay];
            return NSDragOperationCopy;
        } else {
            return NSDragOperationNone;
        }
    }
}

- (void)draggingExited:(id <NSDraggingInfo>)sender
{
    drawForDrop = NO;
    [self setNeedsDisplay];
}

- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
{
    return YES;
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
    if([sender draggingSource] == self){
        // a move - inject selected cells into matrix at mouse point
        [self moveSelectedCellsToIndex:[self injectionIndexForPoint:[sender draggingLocation]]];
    } else {
        // files dragged from other app - paste in at mouse point
        PasteController *pc = [[PasteController alloc] initWithFiles:[[sender draggingPasteboard] propertyListForType:NSFilenamesPboardType] matrix:self injectionIndex:[self injectionIndexForPoint:[sender draggingLocation]]];
        [pc startPaste];
    }
    return YES;
}

- (void)concludeDragOperation:(id <NSDraggingInfo>)sender
{
    drawForDrop = NO;
    [self setNeedsDisplay];
    /*
    NSSound *pasteSound = [[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/system/Volume Mount.aif" byReference:YES];
    if(pasteSound){
        [pasteSound play];
    }
    */
}

#pragma mark -
#pragma mark Sizing

- (void)resizeWithOldSuperviewSize:(NSSize)oldFrameSize
{
    // this gets called when the superview is being resized
    // I want the matrix to grow and shrink based on the width of the superview,
    // with no regard to the height.
    [self adjustSize];
}

- (void)viewDidEndLiveResizeWithWidth:(float)width
{
    // called when the resize is over
    [self adjustSize];
}

- (void)adjustSize
{
    // see if I should change number of columns
    if(_numberOfColumns != [ThumbnailMatrix correctNumberOfColumns:[[self superview] frame].size.width]){
        [self setNumberOfColumns:[ThumbnailMatrix correctNumberOfColumns:[[self superview] frame].size.width]];
    }
    
    int cellWidth = (int)([[self superview] frame].size.width / _numberOfColumns);
    
    [self setCellSize:NSMakeSize(cellWidth,cellWidth)];
    [self sizeToCells];
    [self setNeedsDisplay];
}

#pragma mark -
#pragma mark Convenience

- (NSString *)filePathForCell:(id)cell
{
    return [[cell representedObject] storagePath];
}

- (NSArray *)selectedFiles
{
    NSMutableArray *returnArray = [[NSMutableArray alloc] init];
    PSMActionImageCell *thisCell;
    NSEnumerator *e = [[self selectedCells] objectEnumerator];
    while(thisCell = [e nextObject]){
        [returnArray addObject:[self filePathForCell:thisCell]];
    }
    return returnArray;
}

+ (NSSize) subtractPoints: (NSPoint) a_1 from: (NSPoint) a_2 {
    return NSMakeSize(a_2.x - a_1.x, a_2.y - a_1.y);
}

- (int)injectionIndexForPoint:(NSPoint)point
{
    // receives point in window coords - convert to view coords and find index
    int row, col;
    [self getRow:&row column:&col forPoint:[[[self window] contentView] convertPoint:point toView:self]];
    int injectionIndex = row *_numberOfColumns + col;
    return injectionIndex;
}

- (void)reassignCellsFromArray:(NSMutableArray *)array
{
    [array retain];
    int row = 0, col = 0;
    PSMActionImageCell *thisCell;
    NSEnumerator *e = [array objectEnumerator];
    while(thisCell = [e nextObject]){
        [self putCell:thisCell atRow:row column:col];
        col++;
        if(col >= _numberOfColumns){
            col = 0;
            row++;
        }
    }
    [array release];
}

- (void)selectAll:(id)sender
{
    // pick all cells with actual content
    if(_numImages > 0)
        [self setSelectionFrom:0 to:_numImages-1 anchor:0 highlight:YES];
    [self setNeedsDisplay];
}

+ (int)correctNumberOfColumns:(float)width
{
    int cols = 1;
    while(cols > 0){
        if(width > (cols * 96.0 + 48.0)){
            cols++;
        } else {
            break;
        }
    }
    return cols;
}

@end
